# -*- coding: utf-8 -*-
# Based https://pytorch.org/tutorials/beginner/nlp/advanced_tutorial.html
# Modified by InfinityFuture

import math
import torch
import torch.autograd as autograd
import torch.nn as nn
import torch.optim as optim
from tqdm import tqdm
from sklearn.base import BaseEstimator
import numpy as np

from torch_tagger.rnn_crf import RNN_CRF
from torch_tagger.utils import build_vocabulary, prepare_sequence, batch_flow, DEVICE
from torch_tagger.utils import extrat_entities, pad_seq

class Tagger(BaseEstimator):
    """scikit-learn compatible Tagger"""

    def __init__(
        self,
        embedding_dim=64,
        hidden_dim=32,
        learning_rate=1e-3,
        weight_decay=1e-6,
        epochs=10,
        verbose=1,
        batch_size=32,
        device='auto',
        embedding_dropout_p=0.1,
        rnn_dropout_p=0.1,
        bidirectional=True,
        rnn_type='lstm',
        num_layers=1,
        model=None,
        optimizer=None,
        word_to_ix={},
        ix_to_word={},
        tag_to_ix={},
        ix_to_tag={},
        ):
        """init"""
        self.params = {
            'embedding_dim': embedding_dim,
            'hidden_dim': hidden_dim,
            'learning_rate': learning_rate,
            'weight_decay': weight_decay,
            'epochs': epochs,
            'verbose': verbose,
            'batch_size': batch_size,
            'device': device,
            'embedding_dropout_p': embedding_dropout_p,
            'rnn_dropout_p': rnn_dropout_p,
            'bidirectional': bidirectional,
            'rnn_type': rnn_type,
            'num_layers': num_layers
        }
        self.model = model
        self.optimizer = optimizer
        self.word_to_ix=  word_to_ix
        self.ix_to_word = ix_to_word
        self.tag_to_ix = tag_to_ix
        self.ix_to_tag = ix_to_tag

    def get_params(self, deep=True):
        """Get params for scikit-learn compatible"""
        params = self.params
        if deep:
            params['model'] = self.model.state_dict() if self.model is not None else None
            params['optimizer'] = self.optimizer.state_dict() if self.optimizer is not None else None
            params['word_to_ix'] = self.word_to_ix
            params['ix_to_word'] = self.ix_to_word
            params['tag_to_ix'] = self.tag_to_ix
            params['ix_to_tag'] = self.ix_to_tag
        return params

    def set_params(self, **parameters):
        """Set params for scikit-learn compatible"""
        for k, v in parameters.items():
            if k in self.params:
                self.params[k] = v
        return self

    def __getstate__(self):
        """Get state for pickle"""
        state = {
            'params': self.params,
            'model': self.model.state_dict() if self.model is not None else None,
            'optimizer': self.optimizer.state_dict() if self.optimizer is not None else None,
            'word_to_ix': self.word_to_ix,
            'ix_to_word': self.ix_to_word,
            'tag_to_ix': self.tag_to_ix,
            'ix_to_tag': self.ix_to_tag
        }
        return state
    
    def __setstate__(self, state):
        """Get state for pickle"""
        self.params = state['params']
        if state['model'] is not None:
            self.word_to_ix = state['word_to_ix']
            self.ix_to_word = state['ix_to_word']
            self.tag_to_ix = state['tag_to_ix']
            self.ix_to_tag = state['ix_to_tag']
            self.apply_params()
            self.model.load_state_dict(state['model'])
            self.optimizer.load_state_dict(state['optimizer'])

    def _get_device(self):
        """Get device to predict or train"""
        device = self.params['device']
        if device == 'auto':
            return DEVICE
        elif device in ('gpu', 'cuda'):
            return torch.device('cuda')
        else:
            return torch.device('cpu')
    
    def apply_params(self):
        """Apply params and build RNN-CRF model"""

        embedding_dim = self.params['embedding_dim']
        hidden_dim = self.params['hidden_dim']
        learning_rate = self.params['learning_rate']
        weight_decay = self.params['weight_decay']
        embedding_dropout_p = self.params['embedding_dropout_p']
        rnn_dropout_p = self.params['rnn_dropout_p']
        bidirectional = self.params['bidirectional']
        rnn_type = self.params['rnn_type']
        num_layers = self.params['num_layers']

        word_to_ix = self.word_to_ix
        tag_to_ix = self.tag_to_ix

        model = RNN_CRF(
            len(word_to_ix),
            tag_to_ix,
            embedding_dim,
            hidden_dim,
            num_layers=num_layers,
            bidirectional=bidirectional,
            device=self._get_device(),
            embedding_dropout_p=embedding_dropout_p,
            rnn_dropout_p=rnn_dropout_p,
            rnn_type=rnn_type
        ).to(self._get_device())

        optimizer = optim.Adam(
            model.parameters(),
            lr=learning_rate,
            weight_decay=weight_decay
        )

        self.model = model
        self.optimizer = optimizer
    
    def fit(self, X, y):
        """Fit the model"""

        assert len(X) > 0, 'X must size > 0'
        assert len(y) > 0, 'y must size > 0'
        assert len(X) == len(y), 'X must size equal to y'

        # Autommatic build vocabulary
        if len(self.word_to_ix) <= 0:
            vocabulary = build_vocabulary(X, y)
            self.word_to_ix = vocabulary['word_to_ix']
            self.ix_to_word = vocabulary['ix_to_word']
            self.tag_to_ix = vocabulary['tag_to_ix']
            self.ix_to_tag = vocabulary['ix_to_tag']

        epochs = self.params['epochs']
        verbose = self.params['verbose']
        batch_size = self.params['batch_size']

        word_to_ix = self.word_to_ix
        tag_to_ix = self.tag_to_ix
        ix_to_tag = self.ix_to_tag

        self.apply_params()
        model, optimizer = self.model, self.optimizer

        # Check predictions before training
        with torch.no_grad():
            precheck_sent = prepare_sequence(X[0], word_to_ix)
            precheck_sent = torch.stack([precheck_sent])
            precheck_sent = precheck_sent.to(self._get_device())
            _, p = model(precheck_sent)
            assert len([ix_to_tag[tag_to_ix[t]] for t in y[0]]) == len(p[0]), \
                'checking before training error'

        # Make sure prepare_sequence from earlier in the LSTM section is loaded
        for epoch in range(epochs):
            pbar = range(math.ceil(len(X) / batch_size))
            losses = []
            if verbose > 0:
                pbar = tqdm(pbar, ncols=0)
                pbar.set_description('epoch: {}/{} loss: {:.4f}'.format(epoch + 1, epochs, 0))
            for _ in pbar:

                xb, yb, lb = next(batch_flow(X, y, word_to_ix, tag_to_ix, batch_size))
                xb = xb.to(self._get_device())
                yb = yb.to(self._get_device())
                lb = lb.to(self._get_device())

                # Step 1. Remember that Pytorch accumulates gradients.
                # We need to clear them out before each instance
                model.zero_grad()

                # Step 2. Run our forward pass.
                loss = model.neg_log_likelihood(xb, yb, lb)
                losses.append(loss.item())

                # Step 3. Compute the loss, gradients, and update the parameters by
                # calling optimizer.step()
                loss.backward()
                optimizer.step()
                if verbose > 0:
                    pbar.set_description('epoch: {}/{} loss: {:.4f}'.format(epoch + 1, epochs, np.mean(losses)))
    
    def predict(self, X, batch_size=32, verbose=0):
        """Predict tags"""
        assert len(X) > 0, 'predict empty'
        model = self.model
        word_to_ix = self.word_to_ix
        ix_to_tag = self.ix_to_tag
        
        # Check predictions after training
        ret = []
        with torch.no_grad():

            batch_total = math.ceil(len(X) / batch_size)
            
            pbar = range(batch_total)
            if verbose > 0:
                pbar = tqdm(pbar, ncols=0)
            for i in pbar:
                x_batch = X[i * batch_size : (i + 1) * batch_size]

                max_len = np.max([len(t) for t in x_batch])
                lens = [len(x) for x in x_batch]
                x_batch = [pad_seq(x, max_len) for x in x_batch]
                x_batch = [
                    prepare_sequence(sent, word_to_ix)
                    for sent in x_batch
                ]

                sent = torch.stack(x_batch)
                sent = sent.to(self._get_device())
                _, predicts = model(sent)
                for l, tags in zip(lens, predicts):
                    tags = [ix_to_tag[i] for i in tags[:l]]
                    ret.append(tags)
        return ret

    def score(self, X, y, verbose=0):
        """Calculate NER F1"""
        preds = self.predict(X, verbose=verbose)
        scores = []
        pbar = zip(preds, y)
        if verbose > 0:
            pbar = tqdm(pbar, ncols=0, total=len(y))
        for pred, yy in pbar:
            pset = extrat_entities(pred)
            rset = extrat_entities(yy)
            pset = set(pset)
            rset = set(rset)
            inter = pset.intersection(rset)
            precision = len(inter) / len(pset) if len(pset) > 0 else 1
            recall = len(inter) / len(rset) if len(rset) > 0 else 1
            f1 = 0
            if precision + recall > 0:
                f1 = 2 * ((precision * recall) / (precision + recall))
            scores.append(f1)
        return np.mean(scores)



# -*- coding: utf-8 -*-
# From https://pytorch.org/tutorials/beginner/nlp/advanced_tutorial.html
# Modified by InfinityFuture

from collections import Counter
import numpy as np
import torch

DEVICE = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
START_TAG = '<START>'
STOP_TAG = '<STOP>'
PAD_TAG = '<PAD>'
UNK_TAG = '<UNK>'

def prepare_sequence(seq, to_ix) -> torch.LongTensor:
    """Convert sequence to torch variable"""
    idxs = [
        to_ix[w] if w in to_ix else to_ix[UNK_TAG]
        for w in seq
    ]
    return torch.tensor(idxs, dtype=torch.long)

def default_spliter(x) -> list:
    """Default sentence spliter"""
    return x.split()

def text_reader(path, spliter=default_spliter):
    """Read a text file, and return data
    data should follow this format:
    I want to New York
    O O O CityB CityI
    """
    with open(path, 'r') as fp:
        lines = []
        for l in fp:
            l = l.strip()
            if len(l):
                lines.append(l)
    assert len(lines) > 0, 'text file empty "{}"'.format(path)
    assert len(lines) % 2 == 0, 'text file should have even lines "{}"'.format(path)
    x_data = []
    y_data = []
    for i, l in enumerate(lines):
        if i % 2 == 1:
            line = lines[i - 1]
            tag = l
            line = spliter(line)
            tag = spliter(tag)
            x_data.append(line)
            y_data.append(tag)
            assert len(line) == len(tag), 'line "{}" and "{}" do not match'.format(i - 1, i)
    return x_data, y_data

def build_vocabulary(x_data: list, y_data: list) -> dict:
    """ Use data to build vocabulary"""
    sentence_word = Counter()
    tags_word = Counter()
    for sentence in x_data:
        sentence_word.update(sentence)
    for tags in y_data:
        tags_word.update(tags)

    word_to_ix = {
        PAD_TAG: 0,
        UNK_TAG: 1
    }
    for word in sentence_word.keys():
        ix = len(word_to_ix)
        word_to_ix[word] = ix
    
    ix_to_word = {v: k for k, v in word_to_ix.items()}

    tag_to_ix = {
        PAD_TAG: 0,
        UNK_TAG: 1,
        START_TAG: 2,
        STOP_TAG: 3
    }

    for tag in tags_word.keys():
        ix = len(tag_to_ix)
        tag_to_ix[tag] = ix

    ix_to_tag = { v: k for k, v in tag_to_ix.items()}

    return {
        'word_to_ix': word_to_ix,
        'ix_to_word': ix_to_word,
        'tag_to_ix': tag_to_ix,
        'ix_to_tag': ix_to_tag,
    }

def pad_seq(x: list, max_len: int) -> list:
    """Padding data to max_len length"""
    if len(x) < max_len:
        return x + [PAD_TAG] * (max_len - len(x))
    return x

def batch_flow(x_data: list, y_data: list, word_to_ix: dict, tag_to_ix: dict, batch_size: int= 32):
    """Automatic generate batch data"""
    assert len(x_data) > 0, 'len(x_data) > 0'
    assert len(x_data) == len(y_data), 'len(x_data) == len(y_data), {} != {}'.format(len(x_data), len(y_data))

    x_batch, y_batch, len_batch = [], [], []
    while True:
        if len(x_batch) == batch_size:
            max_len = np.max([len(t) for t in x_batch])
            x_batch = [pad_seq(x, max_len) for x in x_batch]
            y_batch = [pad_seq(y, max_len) for y in y_batch]
            x_batch = [prepare_sequence(x, word_to_ix) for x in x_batch]
            y_batch = [prepare_sequence(y, tag_to_ix) for y in y_batch]

            batches = list(zip(x_batch, y_batch, len_batch))
            batches = sorted(batches, key=lambda x: x[2], reverse=True)
            x_batch = [t[0] for t in batches]
            y_batch = [t[1] for t in batches]
            len_batch = [t[2] for t in batches]

            len_batch = [torch.tensor(t, dtype=torch.long) for t in len_batch]

            yield torch.stack(x_batch), torch.stack(y_batch), torch.stack(len_batch)
            x_batch, y_batch = [], []

        ind = np.random.randint(0, len(x_data))
        x = x_data[ind]
        y = y_data[ind]

        x_batch.append(x)
        y_batch.append(y)
        len_batch.append(len(x))

def sequence_mask(lens: torch.Tensor, max_len: int = None) -> torch.FloatTensor:
    """InfinityFutures:
    This function is copy from:
    https://github.com/epwalsh/pytorch-crf
    The author is epwalsh, and its license is MIT too

    Compute sequence mask.
    Parameters
    ----------
    lens : torch.Tensor
        Tensor of sequence lengths ``[batch_size]``.
    max_len : int, optional (default: None)
        The maximum length (optional).
    Returns
    -------
    torch.ByteTensor
        Returns a tensor of 1's and 0's of size ``[batch_size x max_len]``.
    """
    batch_size = lens.size(0)

    if max_len is None:
        max_len = lens.max().item()

    ranges = torch.arange(0, max_len, device=lens.device).long()
    ranges = ranges.unsqueeze(0).expand(batch_size, max_len)
    ranges = torch.autograd.Variable(ranges)

    lens_exp = lens.unsqueeze(1).expand_as(ranges)
    mask = ranges < lens_exp

    return mask.float()

def extrat_entities(seq: list) -> list:
    """Extract entities from a sequences
    ---
    input: ['B', 'I', 'I', 'O', 'B', 'I']
    output: [(0, 3, ''), (4, 6, '')]
    ---
    input: ['B-loc', 'I-loc', 'I-loc', 'O', 'B-per', 'I-per']
    output: [(0, 3, '-loc'), (4, 6, '-per')]
    """
    ret = []
    start_ind, start_type = -1, None
    for i, s in enumerate(seq):
        if s.startswith('B') or s.startswith('O'):
            if start_ind >= 0:
                ret.append((start_ind, i, start_type))
                start_ind, start_type = -1, None
        if s.startswith('B'):
            start_ind = i
            start_type = s[1:]
    if start_ind >= 0:
        ret.append((start_ind, len(seq), start_type))
        start_ind, start_type = -1, None
    return ret



# -*- coding: utf-8 -*-
# Based https://pytorch.org/tutorials/beginner/nlp/advanced_tutorial.html
# Modified by InfinityFuture

import torch
import torch.autograd as autograd
import torch.nn as nn
import torch.optim as optim

from torch_tagger.utils import START_TAG, STOP_TAG, DEVICE, sequence_mask

def argmax(vec) -> int:
    """return the argmax as a python int"""
    _, idx = torch.max(vec, 1)
    return idx.item()

def log_sum_exp(vec):
    """Compute log sum exp in a numerically stable way for the forward algorithm
    vec dim: [n]
    """
    max_score = vec[0, argmax(vec)]
    max_score_broadcast = max_score.view(1, -1).expand(1, vec.size()[1])
    t = vec - max_score_broadcast
    t = torch.exp(t)
    t = torch.sum(t)
    t = torch.log(t)
    t = max_score + t
    return t

def argmax_batch(vec) -> torch.Tensor:
    """Batch version of argmax"""
    _, idx = torch.max(vec, 1)
    return idx

def log_sum_exp_batch(vec):
    """Batch version of log_sum_exp
    vec dim: [batch_size, n]
    """
    batch_size = vec.size()[0]
    amax = argmax_batch(vec)
    max_score = vec.gather(1, amax.view(-1, 1))
    max_score = max_score.view(1, -1)
    max_score_broadcast = max_score.view(batch_size, -1)
    max_score_broadcast = max_score_broadcast.expand(batch_size, vec.size()[1])
    t = vec - max_score_broadcast
    t = torch.exp(t)
    t = torch.sum(t, -1)
    t = torch.log(t)
    t = max_score + t        
    return t[0]


class RNN_CRF(nn.Module):
    """
    RNN + CRF
    """

    def __init__(
        self,
        vocab_size,
        tag_to_ix,
        embedding_dim,
        hidden_dim,
        num_layers=1,
        bidirectional=True,
        embedding_dropout_p=0.1,
        rnn_dropout_p=0.1,
        rnn_type='lstm',
        device=DEVICE
        ):
        super(RNN_CRF, self).__init__()
        self.embedding_dim = embedding_dim
        self.hidden_dim = hidden_dim
        self.vocab_size = vocab_size
        self.tag_to_ix = tag_to_ix
        self.tagset_size = len(tag_to_ix)
        self.num_layers = num_layers
        self.bidirectional = bidirectional
        self.embedding_dropout_p = embedding_dropout_p
        self.rnn_dropout_p = rnn_dropout_p
        self.device = device
        self.rnn_type = rnn_type

        self.word_embeds = nn.Embedding(vocab_size, embedding_dim)
        if rnn_type == 'lstm':
            self.rnn = nn.LSTM(
                embedding_dim,
                hidden_dim,
                num_layers=num_layers,
                bidirectional=bidirectional
            )
        elif rnn_type == 'gru':
            self.rnn = nn.GRU(
                embedding_dim,
                hidden_dim,
                num_layers=num_layers,
                bidirectional=bidirectional
            )
        else:
            raise Exception('Invalid rnn_type')

        self.embedding_dropout = nn.Dropout(self.embedding_dropout_p)
        self.rnn_dropout = nn.Dropout(self.rnn_dropout_p)

        # Maps the output of the LSTM into tag space.
        self.hidden2tag = nn.Linear(
            hidden_dim * 2 if bidirectional else hidden_dim,
            self.tagset_size
        )

        # Matrix of transition parameters.  Entry i,j is the score of
        # transitioning *to* i *from* j.
        self.transitions = nn.Parameter(
            torch.randn(self.tagset_size, self.tagset_size))

        # These two statements enforce the constraint that we never transfer
        # to the start tag and we never transfer from the stop tag
        self.transitions.data[tag_to_ix[START_TAG], :] = -10000
        self.transitions.data[:, tag_to_ix[STOP_TAG]] = -10000

        # self.hidden = self.init_hidden()

    def init_hidden(self, batch_size):
        """Initilize the hidden for RNN"""
        bidirectional = self.bidirectional
        hidden_dim = self.hidden_dim
        num_layers = self.num_layers
        rnn_type = self.rnn_type
        if rnn_type == 'lstm':
            return (
                torch.randn(
                    2 * num_layers if bidirectional else num_layers,
                    batch_size,
                    hidden_dim
                ).to(self.device),
                torch.randn(
                    2 * num_layers if bidirectional else num_layers,
                    batch_size,
                    hidden_dim
                ).to(self.device),
            )
        elif rnn_type == 'gru':
            return torch.randn(
                2 * num_layers if bidirectional else num_layers,
                batch_size,
                hidden_dim
            ).to(self.device)

    def _forward_alg(self, feats_batch, lengths_batch):
        """Get CRF result"""
        # feats_batch dim: [batch_size, seq_len, target_dim]
        # lengths_batch dim: [batch_size]
        batch_size = feats_batch.size()[0]
        alpha_batch = []
        # Do the forward algorithm to compute the partition function
        init_alphas = torch.full((batch_size, self.tagset_size), -10000.).to(self.device)
        # START_TAG has all of the score.
        init_alphas[:, self.tag_to_ix[START_TAG]] = 0.

        for i in range(batch_size):
            feats = feats_batch[i]
            lengths = lengths_batch[i]

            # Wrap in a variable so that we will get automatic backprop
            forward_var = init_alphas[i]

            # Iterate through the sentence
            for feat in feats[:lengths]:
                alphas_t = []  # The forward tensors at this timestep
                for next_tag in range(self.tagset_size):
                    # broadcast the emission score: it is the same regardless of
                    # the previous tag
                    emit_score = feat[next_tag].view(
                        1, -1).expand(1, self.tagset_size)
                    # the ith entry of trans_score is the score of transitioning to
                    # next_tag from i
                    trans_score = self.transitions[next_tag].view(1, -1)
                    # The ith entry of next_tag_var is the value for the
                    # edge (i -> next_tag) before we do log-sum-exp
                    next_tag_var = forward_var + trans_score + emit_score
                    # The forward variable for this tag is log-sum-exp of all the
                    # scores.
                    alphas_t.append(log_sum_exp(next_tag_var).view(1))
                forward_var = torch.cat(alphas_t).view(1, -1)
                # print('forward_var AAA', forward_var.size())
            terminal_var = forward_var + self.transitions[self.tag_to_ix[STOP_TAG]]
            alpha = log_sum_exp(terminal_var)
            alpha_batch.append(alpha)
        return torch.stack(alpha_batch) / lengths.float()
    
    def _forward_alg_batch(self, feats, lengths, masks):
        """Get CRF result, batch version"""
        # feats dim: [batch_size, seq_len, target_dim]
        # lengths dim: [batch_size]
        # masks dim: [batch_size, seq_len]
        batch_size, seq_len = feats.size()[:2]
        # Do the forward algorithm to compute the partition function
        init_alphas = torch.full((batch_size, self.tagset_size), -10000.).to(self.device)
        # START_TAG has all of the score.
        init_alphas[:, self.tag_to_ix[START_TAG]] = 0.

        # Wrap in a variable so that we will get automatic backprop
        forward_var = init_alphas

        # feats dim: [seq_len, batch_size, target_dim]
        feats = feats.permute(1, 0, 2)
        # Iterate through the sentence
        for i, feat in enumerate(feats):
            alphas_t = []  # The forward tensors at this timestep
            for next_tag in range(self.tagset_size):
                # broadcast the emission score: it is the same regardless of
                # the previous tag
                emit_score = feat[:, next_tag].view(
                    batch_size, -1).expand(batch_size, self.tagset_size)
                # the ith entry of trans_score is the score of transitioning to
                # next_tag from i
                trans_score = self.transitions[next_tag].view(1, -1).expand(
                    batch_size, self.tagset_size)
                trans_score = trans_score * masks[:, i]
                # The ith entry of trans_score is the value for the
                # edge (i -> next_tag) before we do log-sum-exp
                next_tag_var = forward_var + trans_score + emit_score
                # The forward variable for this tag is log-sum-exp of all the
                # scores.
                alphas_t.append(log_sum_exp_batch(next_tag_var))
            forward_var = torch.stack(alphas_t)
            forward_var = forward_var.permute(1, 0)
        terminal_var = forward_var + self.transitions[self.tag_to_ix[STOP_TAG]]
        alpha = log_sum_exp_batch(terminal_var) / lengths.float()
        return alpha

    def _get_rnn_features(self, sentence):
        """Get the output of RNN model"""
        # sentence dim: [batch_size, seq_len]
        batch_size, _ = sentence.size()

        self.hidden = self.init_hidden(batch_size)

        # embeds dim: [batch_size, seq_len, embedding_size]
        embeds = self.word_embeds(sentence)
        embeds = self.embedding_dropout(embeds)
        # embeds dim: [seq_len, batch_size, embedding_size]
        # embeds = embeds.view(seq_len, batch_size, -1)
        embeds = embeds.permute(1, 0, 2)

        # rnn_out dim: [seq_len, batch_size, hidden_dim]
        rnn_out, self.hidden = self.rnn(embeds, self.hidden)
        # When your batch_size > 1, you should use permute not view
        # rnn_out = rnn_out.view(batch_size, seq_len, self.hidden_dim)
        rnn_out = rnn_out.permute(1, 0, 2)
        # rnn_feats dim: [batch_size, seq_len, target_dim]
        rnn_feats = self.hidden2tag(rnn_out)

        return rnn_feats

    def _score_sentence(self, feats_batch, tags_batch, lengths_batch):
        """Get the gold standard of sentences"""
        # Gives the score of a provided tag sequence
        # feats_batch dim: [batch_size, seq_len, target_dim]
        # tags_batch dim: [batch_size, seq_len]
        # lengths_batch dim: [batch_size]
        batch_size = feats_batch.size()[0]
        score_batch = []
        for i in range(batch_size):
            score = torch.zeros(1).to(self.device)
            feats = feats_batch[i]
            tags = tags_batch[i]
            lengths = lengths_batch[i]

            tags = torch.cat([
                torch.tensor([self.tag_to_ix[START_TAG]], dtype=torch.long).to(self.device),
                tags
            ])
            
            for j, feat in enumerate(feats):
                score = score + \
                    self.transitions[tags[j + 1], tags[j]] + feat[tags[j + 1]]
            score = score + self.transitions[self.tag_to_ix[STOP_TAG], tags[-1]]
            score = score / lengths.float()
            score_batch.append(score)
        return torch.cat(score_batch)

    def _score_sentence_batch(self, feats, tags, lengths, masks):
        """Get the gold standard of sentences, batch version"""
        # Gives the score of a provided tag sequence
        # feats dim: [batch_size, seq_len, target_dim]
        # tags dim: [batch_size, seq_len]
        # lengths dim: [batch_size]
        # masks dim: [batch_size, seq_len]
        batch_size = feats.size()[0]

        score = torch.zeros(batch_size).to(self.device)

        start = torch.tensor([
            self.tag_to_ix[START_TAG]
        ], dtype=torch.long).expand(batch_size, 1).to(self.device)

        tags = torch.cat([
            start,
            tags
        ], dim=1)
        
        feats = feats.permute(1, 0, 2)
        for j, feat in enumerate(feats):
            aa, bb = tags[:, j + 1], tags[:, j]
            a = self.transitions[aa, bb]
            a = a * masks[:, j, 0].view(-1)
            b = feat[:, aa]
            score = score + a + b
                
        score = score + self.transitions[self.tag_to_ix[STOP_TAG], tags[:, -1]]
        score = score / lengths.float()

        return score.diag()

    def _viterbi_decode(self, feats_batch):
        """Inference result"""
        # feats_batch dim: [batch_size, seq_len, target_dim]
        batch_size = feats_batch.size()[0]
        path_scores, best_paths = [], []
        for i in range(batch_size):
            feats = feats_batch[i]
            backpointers = []

            # Initialize the viterbi variables in log space
            init_vvars = torch.full((1, self.tagset_size), -10000.).to(self.device)
            init_vvars[0][self.tag_to_ix[START_TAG]] = 0

            # forward_var at step i holds the viterbi variables for step i-1
            forward_var = init_vvars
            for feat in feats:
                bptrs_t = []  # holds the backpointers for this step
                viterbivars_t = []  # holds the viterbi variables for this step

                for next_tag in range(self.tagset_size):
                    # next_tag_var[i] holds the viterbi variable for tag i at the
                    # previous step, plus the score of transitioning
                    # from tag i to next_tag.
                    # We don't include the emission scores here because the max
                    # does not depend on them (we add them in below)
                    next_tag_var = forward_var + self.transitions[next_tag]
                    best_tag_id = argmax(next_tag_var)
                    bptrs_t.append(best_tag_id)
                    viterbivars_t.append(next_tag_var[0][best_tag_id].view(1))
                # Now add in the emission scores, and assign forward_var to the set
                # of viterbi variables we just computed
                forward_var = (torch.cat(viterbivars_t) + feat).view(1, -1)
                backpointers.append(bptrs_t)

            # Transition to STOP_TAG
            terminal_var = forward_var + self.transitions[self.tag_to_ix[STOP_TAG]]
            best_tag_id = argmax(terminal_var)
            path_score = terminal_var[0][best_tag_id]

            # Follow the back pointers to decode the best path.
            best_path = [best_tag_id]
            for bptrs_t in reversed(backpointers):
                best_tag_id = bptrs_t[best_tag_id]
                best_path.append(best_tag_id)
            # Pop off the start tag (we dont want to return that to the caller)
            start = best_path.pop()
            assert start == self.tag_to_ix[START_TAG]  # Sanity check
            best_path.reverse()

            path_scores.append(path_score)
            best_paths.append(best_path)

        return path_scores, best_paths

    def neg_log_likelihood(self, sentence, tags, lengths):
        """Calculate loss"""
        rnn_feats = self._get_rnn_features(sentence)
        batch_size, seq_len, tagset_size = rnn_feats.size()

        masks = sequence_mask(lengths)
        masks = masks.view(batch_size, seq_len, 1)
        masks = masks.expand(batch_size, seq_len, tagset_size)
        rnn_feats = rnn_feats * masks

        rnn_feats = self.rnn_dropout(rnn_feats)

        # forward_score = self._forward_alg(rnn_feats, lengths)
        forward_score = self._forward_alg_batch(rnn_feats, lengths, masks)
        # gold_score = self._score_sentence(rnn_feats, tags, lengths)
        gold_score = self._score_sentence_batch(rnn_feats, tags, lengths, masks)
        loss = torch.mean(forward_score - gold_score)
        return loss

    def forward(self, sentence):  # dont confuse this with _forward_alg above.
        """Main forward function, predict only"""
        # Get the emission scores from the BiLSTM
        rnn_feats = self._get_rnn_features(sentence)
        # Find the best path, given the features.
        score, tag_seq = self._viterbi_decode(rnn_feats)
        return score, tag_seq



# -*- coding: utf-8 -*-

import os
import pickle
from sklearn.model_selection import GridSearchCV

from torch_tagger import Tagger
from torch_tagger.utils import text_reader

CURRENT = os.path.realpath(os.path.dirname(__file__))
PATH = os.path.join(CURRENT, 'train.txt')
x_data, y_data = text_reader(PATH)

def basic_test():
    tag = Tagger(batch_size=8, epochs=500)
    tag.fit(x_data, y_data)

    with open('/tmp/test.pkl', 'wb') as fp:
        pickle.dump(tag, fp)

    tag = None
    tag = pickle.load(open('/tmp/test.pkl', 'rb'))

    score = tag.score(x_data, y_data)
    print('score', score)
    pred = tag.predict(x_data[:32])
    print('truth')
    print(y_data[0])
    print('predict')
    print(pred[0])
    assert str(y_data[0]) == str(pred[0]), 'predict should equal to training data'

def grid_test():
    parameters = {
        'embedding_dim': (16, 32),
        'hidden_dim': (16, 32),
        'bidirectional': (True, False),
        'rnn_type': ('lstm', 'gru'),
        'num_layers': (1, 2)
    }
    tag = Tagger(batch_size=1, epochs=10, verbose=0)
    clf = GridSearchCV(tag, parameters, verbose=2, n_jobs=1)
    clf.fit(x_data, y_data)
    print(clf.cv_results_['mean_test_score'])
    print(clf.best_estimator_)

if __name__ == '__main__':
    # basic_test()
    grid_test()



#!/usr/bin/env python3
#-*- coding:utf-8 -*-

#############################################
# File Name: setup.py
# Author: InfinityFuture
# Mail: infinityfuture@foxmail.com
# Created Time: 2018-09-06 10:00:00
#############################################

import os
from setuptools import setup, find_packages

version = os.path.join(
    os.path.realpath(os.path.dirname(__file__)),
    'version.txt'
)

setup(
    name = 'torch-tagger',
    version = open(version, 'r').read().strip(),
    keywords = ('pip', 'pytorch', 'NER', 'tagger'),
    description = 'NLP tool',
    long_description = 'NLP tool, NER, POS',
    license = 'MIT Licence',

    url = 'https://github.com/infinity-future/torch-tagger',
    author = 'infinityfuture',
    author_email = 'infinityfuture@foxmail.com',

    packages = find_packages(),
    include_package_data = True,
    platforms = 'any',
    install_requires = ['torch', 'tqdm', 'scikit-learn', 'numpy', 'scipy']
)



# -*- coding: utf-8 -*-

import os
import pickle

from torch_tagger import Tagger
from torch_tagger.utils import text_reader


def train_ner():
    x_data, y_data = text_reader('/tmp/train.txt')
    tag = Tagger(batch_size=64, epochs=200, embedding_dim=256, hidden_dim=256)
    tag.fit(x_data, y_data)

    with open('/tmp/test_ner_masks.pkl', 'wb') as fp:
        pickle.dump(tag, fp)

    tag = None
    tag = pickle.load(open('/tmp/test_ner_masks.pkl', 'rb'))

    pred = tag.predict(x_data[:32])
    print('truth')
    print(y_data[0])
    print('predict')
    print(pred[0])
    # assert str(y_data[0]) == str(pred[0]), 'predict should equal to training data'

def test_ner():
    x_data, y_data = text_reader('/tmp/test.txt')
    tag = pickle.load(open('/tmp/test_ner_masks.pkl', 'rb'))

    score = tag.score(x_data, y_data, verbose=1)
    print('score', score)
    pred = tag.predict(x_data[:32])
    print('truth')
    print(y_data[0])
    print('predict')
    print(pred[0])
    # assert str(y_data[0]) == str(pred[0]), 'predict should equal to training data'

if __name__ == '__main__':
    # train_ner()
    test_ner()
